/*********************************************************************************/
/* C++ Code to Generate the Numerator Relationship Matrix (A) which takes        */
/* in two arguments: pedigree file and name of file the matrix will be           */
/* written to. The associated A matrix is saved as row, column, value form.      */
/* The pedigree doesn't have to be sorted and the ID's can be integer or         */
/* alpha-numeric. This code was generated to provide someone a starting          */
/* point of how to generate a relationship with 'real data', instead of          */
/* some simplistic pedigree. Use it at your own discression!!                    */
/*********************************************************************************/
/* To compile type: 'g++ -w -std=c++11 Henderson_1976_A.cpp -o Henderson_1976_A  */
/* To run type: ./Henderson_1976_A pedigree outA                                 */
/*********************************************************************************/
#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <algorithm>
#include <iomanip>
#include <unordered_map>
#include <sstream>

using namespace std;
/*****************************************************************/
/*****   Class Object to Store Information About Each Animal *****/
/*****************************************************************/
class Pedigree_Class
{
private:
    string animal;              /* Original Animal ID */
    string sire;                /* Original Sire ID */
    string dam;                 /* Orignal Dam ID */
    int renum_anim;             /* Renumbered ID */
    int generation;             /* Generation Number */
public:
    Pedigree_Class();
    Pedigree_Class(string anim = "", string si = "", string da = "", int reanim = 0, int gen = 1);
    ~Pedigree_Class();
    /* Functions to retrieve information on an animal */
    string getanim(){return animal;}
    string getsire(){return sire;}
    string getdam(){return dam;}
    int getrenumanim(){return renum_anim;}
    int getgener(){return generation;}
    /* Functions to update information on an animal */
    void Update_Generation(int temp);
    void Update_Renum_Anim(int temp);
    friend bool sortByGen(const Pedigree_Class&, const Pedigree_Class&);        /* Used to sort by generation */
};

Pedigree_Class::Pedigree_Class(){animal = ""; sire = ""; dam = ""; renum_anim = 0; generation = 0;}
Pedigree_Class::Pedigree_Class(string anim, string si, string da, int reanim, int gen)
{
    animal = anim; sire = si; dam = da; renum_anim = reanim; generation = gen;
}
// Destructors
Pedigree_Class::~Pedigree_Class(){};
// Functions
void Pedigree_Class::Update_Generation(int temp){generation = temp;}
void Pedigree_Class::Update_Renum_Anim(int temp){renum_anim = temp;}
// Sort objects
bool sortByGen(const Pedigree_Class &lhs, const Pedigree_Class &rhs) {return lhs.generation > rhs.generation;}

/* Status bar for constructing A Matrix */
static inline void loadbar(unsigned int x, unsigned int n, unsigned int w = 50)
{
    if((x != n) && (x % (n/100+1) != 0)) return;
    float ratio_ =  x/(float)n;
    int c = int(ratio_ * w);
    cout << setw(3) << (int)(ratio_*100) << "% [";
    for (int x=0; x<c; x++) cout << "|";
    for (int x=c; x<w; x++) cout << " ";
    cout << "]\r" << flush;
}
/********************************************/
/*           Start of Program               */
/********************************************/
int main(int argc, char* argv[])
{
    time_t full_begin_time = time(0);
    cout << endl;
    cout <<"+-----+-----+-----+-----+-----+-----+-----+------+-----+" << endl;
    cout <<"|                                                      |" << endl;
    cout <<"|  Henderson (1976) Constructing A                     |" << endl;
    cout <<"|  Author: Jeremy T. Howard                            |" << endl;
    cout <<"|                                                      |" << endl;
    cout <<"+-----+-----+-----+-----+-----+-----+-----+------+-----+" << endl;
    cout <<"| This program is free software: you can redistribute  |" << endl;
    cout <<"| it and/or modify it under the terms of the GNU       |" << endl;
    cout <<"| General Public License as published by the Free      |" << endl;
    cout <<"| Software Foundation, either version 3 of the License,|" << endl;
    cout <<"| or (at your option) any later version.               |" << endl;
    cout <<"+-----+-----+-----+-----+-----+-----+-----+------+-----+" <<endl;
    cout << endl;
    if(argc != 3){cout << "Incorrect Number of parameters specified" << endl; exit (EXIT_FAILURE);}
    /* Generate Output File Name Strings based on Input Parameters */
    string pedigree_file = argv[1];
    string pedigree_sorted = argv[1]; pedigree_sorted += ".SRT";
    string pedigree_file_out = argv[2];
    cout << "- Pedigree file name is '" << pedigree_file << "'." << endl;
    cout << "   - Assume that: " << endl;
    cout << "       - Column 1 is Animal; Column 2 is Sire; Column 3 is Dam." << endl;
    cout << "       - Missing is either a '.', '0' or '-'." << endl;
    cout << "       - No header in file and is space delimeted." << endl;
    cout << "- The sorted pedigree file will be saved to the following file: '" << pedigree_sorted << "'." << endl;
    cout << "- Relationships matrix will be output to the following file: '" << pedigree_file_out << "'." << endl << endl;
    /*************************/
    /* Read in Pedigree file */
    /*************************/
    vector < Pedigree_Class > pedigree;                 /* Save information about animal in vector of "Pedigree_Class" */
    unordered_map <string,int> linker;                  /* Hash-table data structure for faster lookup */
    std:unordered_map<string,int>::const_iterator got;  /* Search if key exists */
    string line,tempanimal,tempsire,tempdam;            /* Initialize temp variables */
    ifstream infile;
    infile.open(pedigree_file.c_str());
    if(infile.fail()){cout << "Pedigree File Doesn't Exist!\n"; exit (EXIT_FAILURE);}
    while (getline(infile,line))
    {
        size_t pos = line.find(" ", 0); tempanimal = line.substr(0,pos); line.erase(0, pos+1);
        /* If have blank lines at the end of file don't add them */
        if(tempanimal != "")
        {
            pos = line.find(" ", 0); tempsire = line.substr(0,pos); line.erase(0, pos+1); tempdam = line;
            /* Weird stuff happens at the end of the row if reading in windows file versus mac file*/
            tempdam.erase(tempdam.find_last_not_of(" \n\r\t")+1);
            tempdam.erase(remove(tempdam.begin(),tempdam.end(),' '),tempdam.end());
            /* if you find one of the possible missing value change it to 0 */
            if(tempsire == "." || tempsire == "-"){tempsire = "0";}
            if(tempdam == "." || tempdam == "-"){tempdam = "0";}
            /* check to make sure animal doesn't already exist in pedigree; if so exit out of program */
            got = linker.find(tempanimal);
            if(got == linker.end())     /* If not found then add */
            {
                /* save in class */
                Pedigree_Class pedigreetemp(tempanimal,tempsire,tempdam,1,1);
                pedigree.push_back(pedigreetemp);
                linker.insert({tempanimal,pedigree.size()-1});
            } else {
                cout << "Exit Program. Animal (" << tempanimal << ") appears more than once in pedigree file!!" << endl;
                exit (EXIT_FAILURE);
            }
        }
    }
    cout << "-Read in pedigree file and perform checks:" << endl;
    cout << "   - Parsed out file into animal, sire and dam." << endl;
    cout << "       - Row 1: - Animal (" << pedigree[0].getanim() << "); Sire (";
    cout << pedigree[0].getsire() << "); Dam (" << pedigree[0].getdam() << ")." << endl;
    cout << "       - Row 2: - Animal (" << pedigree[1].getanim() << "); Sire (";
    cout << pedigree[1].getsire() << "); Dam (" << pedigree[1].getdam() << ")." << endl;
    /* check to make sure all sire and dams are represented as an animal and if missing add (assume both parents are unknown) */
    cout << "   - Make sure all sire or dams appear as an animal." << endl;
    for(int i = 0; i < pedigree.size(); i++)
    {
        if(pedigree[i].getsire() != "0" || pedigree[i].getdam() != "0")
        {
            if(pedigree[i].getsire() != "0")
            {
                /* See if it sire exists as an animal */
                got = linker.find(pedigree[i].getsire());
                if(got == linker.end())     /* If not found then add */
                {
                    cout << "   - Sire: " << pedigree[i].getsire() << " doesn't appear as animal added" << endl;
                    Pedigree_Class pedigreetemp(pedigree[i].getsire(),"0","0",1,1);
                    pedigree.push_back(pedigreetemp);
                    linker.insert({pedigree[i].getsire(),pedigree.size()-1});
                }
            }
            if(pedigree[i].getdam() != "0")
            {
                /* See if it dam exists as an animal */
                got = linker.find(pedigree[i].getdam());
                if(got == linker.end())     /* If not found then add */
                {
                    cout << "   - Dam: " << pedigree[i].getdam() << " doesn't appear as animal added" << endl;
                    Pedigree_Class pedigreetemp(pedigree[i].getdam(),"0","0",1,1);
                    pedigree.push_back(pedigreetemp);
                    linker.insert({pedigree[i].getdam(),(pedigree.size()-1)});
                }
            }
        }
    }
    /* check to make an animal doesn't have a value of 0 */
    for(int i = 0; i < pedigree.size(); i++)
    {
        if(pedigree[i].getanim() == "0")
        {
            cout << endl << "Exit Program. Animal ID can't be '0', line " << i+1 << endl; exit (EXIT_FAILURE);
        }
        if(pedigree[i].getanim() == "")
        {
            cout << endl << "Exit Program. Animal ID can't be '', line " << i+1 << endl; exit (EXIT_FAILURE);
        }
    }
    cout << "   - An animal with an ID of '0' or '' isn't in dataset. " << endl;
    /* check to make sure an animal only is a sire or a dam */
    vector < int > sirecount(pedigree.size(),0);    /* Number of times animal appears as a sire */
    vector < int > damcount(pedigree.size(),0);     /* Number of time animal appears as a dam */
    /* loop through and add a count to a given animal and use linker hash table to look up where sire and dam is at */
    for(int i = 0; i < pedigree.size(); i++)
    {
        if(pedigree[i].getsire() != "0" || pedigree[i].getdam() != "0")
        {
            if(pedigree[i].getsire() != "0"){sirecount[linker[pedigree[i].getsire()]] += 1;}
            if(pedigree[i].getdam() != "0"){damcount[linker[pedigree[i].getdam()]] += 1;}
        }
    }
    for(int i = 0; i < sirecount.size(); i++)
    {
        if(sirecount[i] > 0 && damcount[i] > 0)
        {
            cout << "Exit Program. Animal (" << pedigree[i].getanim() << ") appears as an sire and dam!!" << endl;
            exit (EXIT_FAILURE);
        }
    }
    cout << "   - No animals were found that were a sire and dam in dataset. " << endl;
    cout << "- Finished reading in pedigree (Number of animals " << pedigree.size() << ")." << endl << endl;
    cout << "-Order Pedigree File (i.e. parents come before progeny):" << endl;
    /* Algorithm works by proceeding through the pedigree one animal at a time and all animals begin with generation number = 1. */
    /* Take current generation and increase it by 1 and compare m to the generation number of sire and dam. If it is less than sire */
    /* and dam then change DD to generation 2; This was taken from Dr. Schaeffer's Notes. */
    int orderiteration = 0;
    string done = "NO";
    while(done == "NO")
    {
        orderiteration++; int numberchanged = 0;
        for(int i = 0; i < pedigree.size(); i++)
        {
            if(pedigree[i].getsire() != "0" | pedigree[i].getdam() != "0")
            {
                if(pedigree[i].getsire() != "0")
                {
                    if(pedigree[linker[pedigree[i].getsire()]].getgener() <= pedigree[i].getgener())
                    {
                        pedigree[linker[pedigree[i].getsire()]].Update_Generation(pedigree[i].getgener() + 1); numberchanged++;
                    }
                }
                if(pedigree[i].getdam() != "0")
                {
                    if(pedigree[linker[pedigree[i].getdam()]].getgener() <= pedigree[i].getgener())
                    {
                        pedigree[linker[pedigree[i].getdam()]].Update_Generation(pedigree[i].getgener() + 1); numberchanged++;
                    }
                }
            }
        }
        //if(orderiteration % 1 == 0){cout << "   - " << orderiteration << " (" << numberchanged << ")" << endl;}
        if(numberchanged == 0){done = "YES";}
    }
    cout << "   - Took " << orderiteration << " iterations of the algorithm to figure out order." << endl;
    /* Sort pedigree now by generation so goes from 1:n */
    //for(int i = 0; i < 10; i++) /* Before sort */
    //{
    //   cout<< pedigree[i].getanim()<<" "<<pedigree[i].getsire()<<" "<<pedigree[i].getdam()<<" "<< pedigree[i].getgener()<<endl;
    //}
    sort(pedigree.begin(), pedigree.end(), sortByGen);
    //for(int i = 0; i < 10; i++) /* After sort */
    //{
    //    cout<< pedigree[i].getanim()<<" "<<pedigree[i].getsire()<<" "<<pedigree[i].getdam()<<" "<< pedigree[i].getgener()<<endl;
    //}
    cout << "- Finished Ordering Pedigree." << endl << endl;
    cout << "- Write Sorted Pedigree to file." << endl;
    /* First delete file if it exists */
    fstream checkpedsort; checkpedsort.open(pedigree_sorted.c_str(), std::fstream::out | std::fstream::trunc); checkpedsort.close();
    /* Save as a continuous string and then output */
    stringstream outputstringped(stringstream::out); int outputnumberped = 0;
    for(int i = 0; i < pedigree.size(); i++)
    {
        outputstringped << pedigree[i].getanim()<<" "<<pedigree[i].getsire()<<" "<<pedigree[i].getdam() << endl;
        if(outputnumberped % 1000 == 0)
        {
            /* output pheno pedigree file */
            std::ofstream output10(pedigree_sorted.c_str(), std::ios_base::app | std::ios_base::out);
            output10 << outputstringped.str(); outputstringped.str(""); outputstringped.clear();
        }
        outputnumberped++;
    }
    std::ofstream output10(pedigree_sorted.c_str(), std::ios_base::app | std::ios_base::out);
    output10 << outputstringped.str(); outputstringped.str(""); outputstringped.clear();
    cout << "- Finished Writing Pedigree File." << endl << endl;
    cout << "- Begin constructing Relationship Matrix." << endl;
    time_t A_begin_time = time(0);
    /* Now set renumber ID so they go from 1:number of animals and update linker file since they were sorted*/
    linker.clear(); /* First clear old linker file */
    for(int i = 0; i < pedigree.size(); i++)
    {
        pedigree[i].Update_Renum_Anim(i+1);         /* sets ids to go from 1:n */
        linker.insert({pedigree[i].getanim(),i});   /* updates linker file */
    }
    /* now generate animal, sire and dam vectors to use to generate A based on renum IDs */
    vector < int > renum_animal(pedigree.size(),0);
    vector < int > renum_sire(pedigree.size(),0);
    vector < int > renum_dam(pedigree.size(),0);
    /* Now set renumber ID */
    for(int i = 0; i < pedigree.size(); i++)
    {
        renum_animal[i] = pedigree[i].getrenumanim();
        if(pedigree[i].getsire() != "0" | pedigree[i].getdam() != "0")
        {
            if(pedigree[i].getsire() != "0"){renum_sire[i] = pedigree[linker[pedigree[i].getsire()]].getrenumanim();}
            if(pedigree[i].getdam() != "0"){renum_dam[i] = pedigree[linker[pedigree[i].getdam()]].getrenumanim();}
        }
    }
    /* What pedigree file looks like now */
    //for(int i = 0; i < 20; i++){cout<<renum_animal[i]<<" "<<renum_sire[i]<<" "<<renum_dam[i]<< endl;}
    /* only store lower dimension area though; ensure when index always based on lower triangle */
    /* initialzie 2-D vector to 0 */
    vector< vector < double >> Amatrix;
    for(int i = 0; i < pedigree.size(); i++)
    {
        vector < double > row(i+1,0); Amatrix.push_back(row);
    }
    /* Construct Full A Matrix */
    for(int i = 0; i < renum_animal.size(); i++)
    {
        if (renum_sire[i] != 0 && renum_dam[i] != 0)
        {
            for (int j = 0; j < i; j++)
            {
                if(j <= (renum_sire[i]-1) && j <= (renum_dam[i]-1)){
                    Amatrix[i][j] = 0.5 * (Amatrix[(renum_sire[i]-1)][j] + Amatrix[(renum_dam[i]-1)][j]);
                } else if (j > (renum_sire[i]-1) && j <= (renum_dam[i]-1)){
                    Amatrix[i][j] = 0.5 * (Amatrix[j][(renum_sire[i]-1)] + Amatrix[(renum_dam[i]-1)][j]);
                } else if (j <= (renum_sire[i]-1) && j > (renum_dam[i]-1)){
                    Amatrix[i][j] = 0.5 * (Amatrix[(renum_sire[i]-1)][j] + Amatrix[j][(renum_dam[i]-1)]);
                } else if (j > (renum_sire[i]-1) && j > (renum_dam[i]-1)){
                    Amatrix[i][j] = 0.5 * (Amatrix[j][(renum_sire[i]-1)] + Amatrix[j][(renum_dam[i]-1)]);
                } else {cout << "Shouldn't be here" << endl; exit (EXIT_FAILURE);}
            }
            if((renum_sire[i]-1) < (renum_dam[i]-1)) {
                Amatrix[(renum_animal[i]-1)][(renum_animal[i]-1)] = 1 + 0.5 * Amatrix[(renum_dam[i]-1)][(renum_sire[i]-1)];
            } else if ((renum_sire[i]-1) > (renum_dam[i]-1)) {
                Amatrix[(renum_animal[i]-1)][(renum_animal[i]-1)] = 1 + 0.5 * Amatrix[(renum_sire[i]-1)][(renum_dam[i]-1)];
            } else {cout << "Shouldn't be here" << endl; exit (EXIT_FAILURE);}
        }
        if (renum_sire[i] != 0 && renum_dam[i] == 0)
        {
            for (int j = 0; j < i; j++)
            {
                if(j <= (renum_sire[i]-1)){
                    Amatrix[i][j] = 0.5 * (Amatrix[(renum_sire[i]-1)][j]);
                } else if (j > (renum_sire[i]-1)){
                    Amatrix[i][j] = 0.5 * (Amatrix[j][(renum_sire[i]-1)]);
                } else {cout << "Shouldn't be here" << endl; exit (EXIT_FAILURE);}
            }
            Amatrix[(renum_animal[i]-1)][(renum_animal[i]-1)] = 1;
        }
        if (renum_sire[i] == 0 && renum_dam[i] != 0)
        {
            for (int j = 0; j < i; j++)
            {
                if(j <= (renum_dam[i]-1)){
                    Amatrix[i][j] = 0.5 * (Amatrix[(renum_dam[i]-1)][j]);
                } else if (j > (renum_dam[i]-1)){
                    Amatrix[i][j] = 0.5 * (Amatrix[j][(renum_dam[i]-1)]);
                } else {cout << "Shouldn't be here" << endl; exit (EXIT_FAILURE);}
            }
            Amatrix[(renum_animal[i]-1)][(renum_animal[i]-1)] = 1;
        }
        if (renum_sire[i] == 0 && renum_dam[i] == 0)
        {
            Amatrix[(renum_animal[i]-1)][(renum_animal[i]-1)] = 1;
        }
        //for(int ind1 = 0; ind1 < pedigree.size(); ind1++)
        //{
        //    for(int ind2 = 0; ind2 < pedigree.size(); ind2++)
        //    {
        //        if(ind2 <= ind1){
        //            cout << Amatrix[ind1][ind2] << " ";
        //        } else {ind2 = pedigree.size();}
        //    }
        //    cout << endl;
        //}
        //cout << endl;
    }
    time_t A_end_time = time(0);
    cout << "- Finished Constructing A Matrix (Took: ";
    cout << difftime(A_end_time,A_begin_time) << " seconds)" << endl << endl;
    cout << "- Write A to a file." << endl;
    fstream checkRel; checkRel.open(pedigree_file_out.c_str(), std::fstream::out | std::fstream::trunc); checkRel.close();
    /* Save as a continuous string and then output */
    stringstream outputstring(stringstream::out); int outputnumber = 0;
    /* output diagonal and lower portion of relationship matrix into triplet form */
    for(int ind1 = 0; ind1 < pedigree.size(); ind1++)
    {
        for(int ind2 = 0; ind2 < pedigree.size(); ind2++)
        {
            if(ind2 <= ind1)
            {
                outputstring << pedigree[ind1].getanim() << " "<<pedigree[ind2].getanim()<<" "<< Amatrix[ind1][ind2] << endl;
                if(outputnumber % 100000 == 0)
                {
                    /* output pheno pedigree file */
                    std::ofstream output9(pedigree_file_out.c_str(), std::ios_base::app | std::ios_base::out);
                    output9 << outputstring.str(); outputstring.str(""); outputstring.clear();
                }
                outputnumber++;
            } else {ind2 = pedigree.size();}
        } /* Finish loop across ind2 */
        loadbar(ind1+1,pedigree.size());
    } /* Finish loop across ind1 */
    cout << endl;
    /* output rest of them */
    std::ofstream output9(pedigree_file_out.c_str(), std::ios_base::app | std::ios_base::out);
    output9 << outputstring.str(); outputstring.str(""); outputstring.clear();
    cout << "- Finished Writing A to a File." << endl << endl;
    time_t full_end_time = time(0);
    cout << "- Finished Generating A Matrix (Took: "<<difftime(full_end_time,full_begin_time)<<" seconds)"<<endl<<endl;
}
